@@ -5,7 +5,7 @@ module MCollective
     class Client
       attr_accessor :timeout, :verbose, :filter, :config, :progress, :ttl, :reply_to
       attr_reader :client, :stats, :ddl, :agent, :limit_targets, :limit_method, :output_format, :batch_size, :batch_sleep_time, :batch_mode
-      attr_reader :discovery_options, :discovery_method, :limit_seed
+      attr_reader :discovery_options, :discovery_method, :default_discovery_method, :limit_seed
 
       @@initial_options = nil
 
@@ -25,7 +25,12 @@ module MCollective
           initial_options = Marshal.load(@@initial_options)
 
         else
-          oparser = MCollective::Optionparser.new({:verbose => false, :progress_bar => true, :mcollective_limit_targets => false, :batch_size => nil, :batch_sleep_time => 1}, "filter")
+          oparser = MCollective::Optionparser.new({ :verbose => false,
+                                                    :progress_bar => true,
+                                                    :mcollective_limit_targets => false,
+                                                    :batch_size => nil,
+                                                    :batch_sleep_time => 1 },
+                                                  "filter")
 
           initial_options = oparser.parse do |parser, opts|
             if block_given?
@@ -39,12 +44,15 @@ module MCollective
         end
 
         @initial_options = initial_options
+
+        @config = initial_options[:config]
+        @client = MCollective::Client.new(@initial_options)
+
         @stats = Stats.new
         @agent = agent
         @timeout = initial_options[:timeout] || 5
         @verbose = initial_options[:verbose]
         @filter = initial_options[:filter] || Util.empty_filter
-        @config = initial_options[:config]
         @discovered_agents = nil
         @progress = initial_options[:progress_bar]
         @limit_targets = initial_options[:mcollective_limit_targets]
@@ -53,23 +61,28 @@ module MCollective
         @output_format = initial_options[:output_format] || :console
         @force_direct_request = false
         @reply_to = initial_options[:reply_to]
-        @discovery_method = initial_options[:discovery_method] || Config.instance.default_discovery_method
+        @discovery_method = initial_options[:discovery_method]
+        if !@discovery_method
+          @discovery_method = Config.instance.default_discovery_method
+          @default_discovery_method = true
+        else
+          @default_discovery_method = false
+        end
         @discovery_options = initial_options[:discovery_options] || []
         @force_display_mode = initial_options[:force_display_mode] || false
 
-        @batch_size = Integer(initial_options[:batch_size] || 0)
+        @batch_size = initial_options[:batch_size] || 0
         @batch_sleep_time = Float(initial_options[:batch_sleep_time] || 1)
-        @batch_mode = @batch_size > 0
+        @batch_mode = determine_batch_mode(@batch_size)
 
         agent_filter agent
 
-        @client = MCollective::Client.new(@config)
-        @client.options = initial_options
-
-        @discovery_timeout = discovery_timeout
+        @discovery_timeout = @initial_options.fetch(:disctimeout, nil) || Config.instance.discovery_timeout
 
         @collective = @client.collective
         @ttl = initial_options[:ttl] || Config.instance.ttl
+        @publish_timeout = initial_options[:publish_timeout] || Config.instance.publish_timeout
+        @threaded = initial_options[:threaded] || Config.instance.threaded
 
         # if we can find a DDL for the service override
         # the timeout of the client so we always magically
@@ -88,7 +101,7 @@ module MCollective
         # DDLs are required, failure to find a DDL is fatal
         @ddl = DDL.new(agent)
         @stats.ddl = @ddl
-        @timeout = @ddl.meta[:timeout] + @discovery_timeout if @timeout == 5
+        @timeout = @ddl.meta[:timeout] + discovery_timeout if @timeout == 5
 
         # allows stderr and stdout to be overridden for testing
         # but also for web apps that might not want a bunch of stuff
@@ -230,6 +243,11 @@ module MCollective
 
         validate_request(action, args)
 
+        # TODO(ploubser): The logic here seems poor. It implies that it is valid to
+        # pass arguments where batch_mode is set to false and batch_mode > 0.
+        # If this is the case we completely ignore the supplied value of batch_mode
+        # and do our own thing.
+
         # if a global batch size is set just use that else set it
         # in the case that it was passed as an argument
         batch_mode = args.include?(:batch_size) || @batch_mode
@@ -239,7 +257,7 @@ module MCollective
         # if we were given a batch_size argument thats 0 and batch_mode was
         # determined to be on via global options etc this will allow a batch_size
         # of 0 to disable or batch_mode for this call only
-        batch_mode = (batch_mode && Integer(batch_size) > 0)
+        batch_mode = determine_batch_mode(batch_size)
 
         # Handle single target requests by doing discovery and picking
         # a random node.  Then do a custom request specifying a filter
@@ -333,11 +351,40 @@ module MCollective
       end
 
       def discovery_timeout
-        return @initial_options[:disctimeout] if @initial_options[:disctimeout]
+        return @discovery_timeout if @discovery_timeout
         return @client.discoverer.ddl.meta[:timeout]
       end
 
+      def discovery_timeout=(timeout)
+        @discovery_timeout = Float(timeout)
+
+        # we calculate the overall timeout from the DDL of the agent and
+        # the supplied discovery timeout unless someone specifically
+        # specifies a timeout to the constructor
+        #
+        # But if we also then specifically set a discovery_timeout on the
+        # agent that has to override the supplied timeout so we then
+        # calculate a correct timeout based on DDL timeout and the
+        # supplied discovery timeout
+        @timeout = @ddl.meta[:timeout] + discovery_timeout
+      end
+
+      # Sets the discovery method.  If we change the method there are a
+      # number of steps to take:
+      #
+      #  - set the new method
+      #  - if discovery options were provided, re-set those to initially
+      #    provided ones else clear them as they might now apply to a
+      #    different provider
+      #  - update the client options so it knows there is a new discovery
+      #    method in force
+      #  - reset discovery data forcing a discover on the next request
+      #
+      # The remaining item is the discovery timeout, we leave that as is
+      # since that is the user supplied timeout either via initial options
+      # or via specifically setting it on the client.
       def discovery_method=(method)
+        @default_discovery_method = false
         @discovery_method = method
 
         if @initial_options[:discovery_options]
@@ -347,7 +394,7 @@ module MCollective
         end
 
         @client.options = options
-        @discovery_timeout = discovery_timeout
+
         reset
       end
 
@@ -358,7 +405,7 @@ module MCollective
 
       # Sets the class filter
       def class_filter(klass)
-        @filter["cf_class"] << klass
+        @filter["cf_class"] = @filter["cf_class"] | [klass]
         @filter["cf_class"].compact!
         reset
       end
@@ -370,10 +417,10 @@ module MCollective
 
         if value.nil?
           parsed = Util.parse_fact_string(fact)
-          @filter["fact"] << parsed unless parsed == false
+          @filter["fact"] = @filter["fact"] | [parsed] unless parsed == false
         else
           parsed = Util.parse_fact_string("#{fact}#{operator}#{value}")
-          @filter["fact"] << parsed unless parsed == false
+          @filter["fact"] = @filter["fact"] | [parsed] unless parsed == false
         end
 
         @filter["fact"].compact!
@@ -382,21 +429,21 @@ module MCollective
 
       # Sets the agent filter
       def agent_filter(agent)
-        @filter["agent"] << agent
+        @filter["agent"] = @filter["agent"] | [agent]
         @filter["agent"].compact!
         reset
       end
 
       # Sets the identity filter
       def identity_filter(identity)
-        @filter["identity"] << identity
+        @filter["identity"] = @filter["identity"] | [identity]
         @filter["identity"].compact!
         reset
       end
 
       # Set a compound filter
       def compound_filter(filter)
-        @filter["compound"] <<  Matcher.create_compound_callstack(filter)
+        @filter["compound"] = @filter["compound"] |  [Matcher.create_compound_callstack(filter)]
         reset
       end
 
@@ -462,24 +509,8 @@ module MCollective
             @discovered_agents = hosts
             @force_direct_request = true
 
-          # if an identity filter is supplied and it is all strings no regex we can use that
-          # as discovery data, technically the identity filter is then redundant if we are
-          # in direct addressing mode and we could empty it out but this use case should
-          # only really be for a few -I's on the CLI
-          #
-          # For safety we leave the filter in place for now, that way we can support this
-          # enhancement also in broadcast mode.
-          #
-          # This is only needed for the 'mc' discovery method, other methods might change
-          # the concept of identity to mean something else so we should pass the full
-          # identity filter to them
-          elsif options[:filter]["identity"].size > 0 && @discovery_method == "mc"
-            regex_filters = options[:filter]["identity"].select{|i| i.match("^\/")}.size
-
-            if regex_filters == 0
-              @discovered_agents = options[:filter]["identity"].clone
-              @force_direct_request = true if Config.instance.direct_addressing
-            end
+          else
+            identity_filter_discovery_optimization
           end
         end
 
@@ -510,10 +541,11 @@ module MCollective
           # and if we're configured to use the first found hosts as the
           # limit method then pass in the limit thus minimizing the amount
           # of work we do in the discover phase and speeding it up significantly
+          filter = @filter.merge({'collective' => @collective})
           if @limit_method == :first and @limit_targets.is_a?(Fixnum)
-            @discovered_agents = @client.discover(@filter, discovery_timeout, @limit_targets)
+            @discovered_agents = @client.discover(filter, discovery_timeout, @limit_targets)
           else
-            @discovered_agents = @client.discover(@filter, discovery_timeout)
+            @discovered_agents = @client.discover(filter, discovery_timeout)
           end
 
           @stderr.puts(@discovered_agents.size) if verbose
@@ -532,7 +564,7 @@ module MCollective
       # Provides a normal options hash like you would get from
       # Optionparser
       def options
-        {:disctimeout => @discovery_timeout,
+        {:disctimeout => discovery_timeout,
          :timeout => @timeout,
          :verbose => @verbose,
          :filter => @filter,
@@ -542,7 +574,9 @@ module MCollective
          :discovery_method => @discovery_method,
          :discovery_options => @discovery_options,
          :force_display_mode => @force_display_mode,
-         :config => @config}
+         :config => @config,
+         :publish_timeout => @publish_timeout,
+         :threaded => @threaded}
       end
 
       # Sets the collective we are communicating with
@@ -556,7 +590,13 @@ module MCollective
 
       # Sets and sanity checks the limit_targets variable
       # used to restrict how many nodes we'll target
+      # Limit targets can be reset by passing nil or false
       def limit_targets=(limit)
+        if !limit
+          @limit_targets = nil
+          return
+        end
+
         if limit.is_a?(String)
           raise "Invalid limit specified: #{limit} valid limits are /^\d+%*$/" unless limit =~ /^\d+%*$/
 
@@ -582,10 +622,14 @@ module MCollective
 
       # Sets the batch size, if the size is set to 0 that will disable batch mode
       def batch_size=(limit)
-        raise "Can only set batch size if direct addressing is supported" unless Config.instance.direct_addressing
+        unless Config.instance.direct_addressing
+          raise "Can only set batch size if direct addressing is supported"
+        end
 
-        @batch_size = Integer(limit)
-        @batch_mode = @batch_size > 0
+        validate_batch_size(limit)
+
+        @batch_size = limit
+        @batch_mode = determine_batch_mode(@batch_size)
       end
 
       def batch_sleep_time=(time)
@@ -686,6 +730,8 @@ module MCollective
       def fire_and_forget_request(action, args, filter=nil)
         validate_request(action, args)
 
+        identity_filter_discovery_optimization
+
         req = new_request(action.to_s, args)
 
         filter = options[:filter] unless filter
@@ -693,7 +739,34 @@ module MCollective
         message = Message.new(req, nil, {:agent => @agent, :type => :request, :collective => @collective, :filter => filter, :options => options})
         message.reply_to = @reply_to if @reply_to
 
-        return @client.sendreq(message, nil)
+        if @force_direct_request || @client.discoverer.force_direct_mode?
+          message.discovered_hosts = discover.clone
+          message.type = :direct_request
+        end
+
+        client.sendreq(message, nil)
+      end
+
+      # if an identity filter is supplied and it is all strings no regex we can use that
+      # as discovery data, technically the identity filter is then redundant if we are
+      # in direct addressing mode and we could empty it out but this use case should
+      # only really be for a few -I's on the CLI
+      #
+      # For safety we leave the filter in place for now, that way we can support this
+      # enhancement also in broadcast mode.
+      #
+      # This is only needed for the 'mc' discovery method, other methods might change
+      # the concept of identity to mean something else so we should pass the full
+      # identity filter to them
+      def identity_filter_discovery_optimization
+        if options[:filter]["identity"].size > 0 && @discovery_method == "mc"
+          regex_filters = options[:filter]["identity"].select{|i| i.match("^\/")}.size
+
+          if regex_filters == 0
+            @discovered_agents = options[:filter]["identity"].clone
+            @force_direct_request = true if Config.instance.direct_addressing
+          end
+        end
       end
 
       # Calls an agent in a way very similar to call_agent but it supports batching
@@ -706,8 +779,8 @@ module MCollective
       def call_agent_batched(action, args, opts, batch_size, sleep_time, &block)
         raise "Batched requests requires direct addressing" unless Config.instance.direct_addressing
         raise "Cannot bypass result processing for batched requests" if args[:process_results] == false
+        validate_batch_size(batch_size)
 
-        batch_size = Integer(batch_size)
         sleep_time = Float(sleep_time)
 
         Log.debug("Calling #{agent}##{action} in batches of #{batch_size} with sleep time of #{sleep_time}")
@@ -729,10 +802,22 @@ module MCollective
             @stdout.print twirl.twirl(respcount, discovered.size)
           end
 
-          @stats.requestid = nil
+          if (batch_size =~ /^(\d+)%$/)
+            # determine batch_size as a percentage of the discovered array's size
+            batch_size = (discovered.size / 100.0 * Integer($1)).ceil
+          else
+            batch_size = Integer(batch_size)
+          end
 
-          discovered.in_groups_of(batch_size) do |hosts, last_batch|
-            message = Message.new(req, nil, {:agent => @agent, :type => :direct_request, :collective => @collective, :filter => opts[:filter], :options => opts})
+          @stats.requestid = nil
+          processed_nodes = 0
+
+          discovered.in_groups_of(batch_size) do |hosts|
+            message = Message.new(req, nil, {:agent => @agent,
+                                             :type => :direct_request,
+                                             :collective => @collective,
+                                             :filter => opts[:filter],
+                                             :options => opts})
 
             # first time round we let the Message object create a request id
             # we then re-use it for future requests to keep auditing sane etc
@@ -755,13 +840,20 @@ module MCollective
               end
             end
 
+            if @initial_options[:sort]
+              results.sort!
+            end
+
             @stats.noresponsefrom.concat @client.stats[:noresponsefrom]
             @stats.responses += @client.stats[:responses]
             @stats.blocktime += @client.stats[:blocktime] + sleep_time
             @stats.totaltime += @client.stats[:totaltime]
             @stats.discoverytime += @client.stats[:discoverytime]
 
-            sleep sleep_time unless last_batch
+            processed_nodes += hosts.length
+            if (discovered.length > processed_nodes)
+              sleep sleep_time
+            end
           end
 
           @stats.aggregate_summary = aggregate.summarize if aggregate
@@ -845,6 +937,10 @@ module MCollective
             end
           end
 
+          if @initial_options[:sort]
+            results.sort!
+          end
+
           @stats.aggregate_summary = aggregate.summarize if aggregate
           @stats.aggregate_failures = aggregate.failed if aggregate
           @stats.client_stats = @client.stats
@@ -894,35 +990,46 @@ module MCollective
         result = rpc_result_from_reply(@agent, action, resp)
         aggregate = aggregate_reply(result, aggregate) if aggregate
 
-        if resp[:body][:statuscode] == 0 || resp[:body][:statuscode] == 1
-          @stats.ok if resp[:body][:statuscode] == 0
-          @stats.fail if resp[:body][:statuscode] == 1
-          @stats.time_block_execution :start
+        @stats.ok if resp[:body][:statuscode] == 0
+        @stats.fail if resp[:body][:statuscode] != 0
+        @stats.time_block_execution :start
 
-          case block.arity
-            when 1
-              block.call(resp)
-            when 2
-              block.call(resp, result)
-          end
+        case block.arity
+          when 1
+            block.call(resp)
+          when 2
+            block.call(resp, result)
+        end
 
-          @stats.time_block_execution :end
-        else
-          @stats.fail
+        @stats.time_block_execution :end
 
-          case resp[:body][:statuscode]
-            when 2
-              raise UnknownRPCAction, resp[:body][:statusmsg]
-            when 3
-              raise MissingRPCData, resp[:body][:statusmsg]
-            when 4
-              raise InvalidRPCData, resp[:body][:statusmsg]
-            when 5
-              raise UnknownRPCError, resp[:body][:statusmsg]
+        return aggregate
+      end
+
+      private
+
+      def determine_batch_mode(batch_size)
+        if (batch_size != 0 && batch_size != "0")
+          return true
+        end
+
+        return false
+      end
+
+      # Validate the bach_size based on the following criteria
+      # batch_size is percentage string and it's more than 0 percent
+      # batch_size is a string of digits
+      # batch_size is of type Integer
+      def validate_batch_size(batch_size)
+        if (batch_size.is_a?(Integer))
+          return
+        elsif (batch_size.is_a?(String))
+          if ((batch_size =~ /^(\d+)%$/ && Integer($1) != 0) || batch_size =~ /^(\d+)$/)
+            return
           end
         end
 
-        return aggregate
+        raise("batch_size must be an integer or match a percentage string (e.g. '24%'")
       end
     end
   end
